/*
-------------------------------------------------------------

                   Multiple Decision Diagrams

			Implementation

-------------------------------------------------------------
*/


/*
-------------------------------------------------------------
	Module 		    :		mdd.c
	Date		    :		10/02/97
	Version		    :		1.21
	Author		    :		Dorel Marius Bozga
-------------------------------------------------------------
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mdd.h"

/*
-------------------------------------------------------------
	Constants and Macros
-------------------------------------------------------------
*/

#define mdd_chain 		16	
#define mdd_dep 		2
#define mdd_growth              256

#define mdd_mask_index		0x0000FFFF
#define mdd_mask_value		0x00FF0000

#define mdd_flag_one 		0x01000000
#define mdd_flag_zero  		0x02000000
#define mdd_flag_lock		0x04000000
#define mdd_flag_mark  		0x08000000

#define mdd_flag_busy   	0x10000000
#define mdd_flag_user1  	0x20000000
#define mdd_flag_user2 		0x40000000
#define mdd_flag_user3 		0x80000000

#define mdd_flag_set(x,f)   	(x->code |= f)
#define mdd_flag_reset(x,f) 	(x->code &= ~f)
#define mdd_flag_test(x,f)  	(x->code & f)	

#define mdd_get_index(x)	(x->code & mdd_mask_index)
#define mdd_set_index(x,i)	(x->code |= i)

#define mdd_get_value(x)	((x->code & mdd_mask_value) >> 16)
#define mdd_set_value(x,v)	(x->code |= (v << 16))

#define mdd_is_one(x)		mdd_flag_test(x,mdd_flag_one)
#define mdd_is_zero(x)		mdd_flag_test(x,mdd_flag_zero)
#define mdd_is_constant(x)      (mdd_is_one(x) || mdd_is_zero(x))

#define mdd_message(m)		fprintf(stderr,"\tmdd : %s\n",m)
#define mdd_abort()		exit(0)

#define mdd_array_alloc(T,n,a)	a=(T*) malloc(sizeof(T) * n)
#define mdd_array_free(a)	free(a);

/*
-------------------------------------------------------------
	MDD Local Functions
-------------------------------------------------------------
*/

	void mdd_flag_recset();

	int  mdd_manager_garbage_collection();
	mdd* mdd_manager_search_or_insert();

	mdd* mdd_basic_exist();
	mdd* mdd_basic_and_exist();
	mdd* mdd_basic_substitute();
	void mdd_basic_print();
	void mdd_basic_dump();
	double mdd_basic_onset_size();
	void mdd_basic_size();

	int  mdd_next_prime();

/*
-------------------------------------------------------------
        MDD Manager Allocation and Destruction
-------------------------------------------------------------
*/

mdd_manager*
mdd_manager_create(vars, mem)
int vars, mem;
/*
Create and initialize and mdd node manager.  The arguments are  
the number of  variables (vars), which must be in the interval 
[1,65535] and the MB of memory allowed (mem), which must be in 
the interval [1,512].
*/
   	{
   	mdd_manager* manager;
   	int i;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_create");
#endif
   	/* test the arguments */
   	if (vars <= 0 || vars > 65535){
   		mdd_message("mdd_create_manager : invalid number of variables");
   		mdd_abort();
  		}
   	if (mem <= 0 || mem > 512){
		mdd_message("mdd_create_manager : invalid memory size");
   		mdd_abort();
  		}

   	/* allocate the manager and then initialize manager structure */
   	manager = (mdd_manager*) malloc(sizeof(mdd_manager));

      	/* initialize memory fields */
   	manager->allowed = 1048576 * mem;

   	/* allocate and initialize variable fields */
   	manager->variable.number = vars;
	mdd_array_alloc(int,vars,manager->variable.range);
	mdd_array_alloc(int,vars,manager->variable.buffer.low);
	mdd_array_alloc(int,vars,manager->variable.buffer.high);
   	for(i=0; i<vars; i++)
   		manager->variable.range[i] = 
		manager->variable.buffer.low[i] = 
		manager->variable.buffer.high[i] =
		0;
   
   	/* allocate and initialize node hash table */
	manager->hashtable.capacity = manager->allowed / sizeof(mdd);
	manager->hashtable.nodes = 0;
   	manager->hashtable.size = mdd_next_prime
   		(manager->hashtable.capacity /  mdd_chain);
	mdd_array_alloc(mdd*,manager->hashtable.size,manager->hashtable.entry);
   	for(i=0;i<manager->hashtable.size;i++)
   		manager->hashtable.entry[i] = NULL;

	/* initialize node table - use static allocation */
	mdd_array_alloc(mdd,manager->hashtable.capacity,manager->table.nodes);
	if (manager->table.nodes == NULL){
	        mdd_message("mdd_create_manager : insufficient memory");
		mdd_abort();
	        }
	for(i=0; i<manager->hashtable.capacity; i++)
	        mdd_flag_reset((manager->table.nodes+i),mdd_flag_busy);
	manager->table.crt = manager->table.nodes;
   	
   	/* allocate and initialize operations caches */
   	manager->cache.size = mdd_next_prime
		(manager->hashtable.size / mdd_dep); 
   	for(i=0; i<3; i++){
      		mdd_array_alloc(mdd*,manager->cache.size,manager->cache.and[i]);
		mdd_array_alloc(mdd*,manager->cache.size,manager->cache.andex[i]);
   		mdd_array_alloc(mdd*,manager->cache.size,manager->cache.or[i]);
   		if (i != 2){
   			mdd_array_alloc(mdd*,manager->cache.size,manager->cache.not[i]);
  			mdd_array_alloc(mdd*,manager->cache.size,manager->cache.exist[i]);
      			mdd_array_alloc(mdd*,manager->cache.size,manager->cache.subst[i]);
      			}
      		}
      	for(i=0;i<manager->cache.size;i++)
      		manager->cache.and[0][i] = manager->cache.and[1][i] = 
      		manager->cache.and[2][i] = manager->cache.or[0][i]  =
		manager->cache.or[1][i]  = manager->cache.or[2][i]  =
		manager->cache.not[0][i] = manager->cache.not[1][i] =
		manager->cache.exist[0][i] = manager->cache.exist[1][i] =
		manager->cache.subst[0][i] = manager->cache.subst[1][i] =
		manager->cache.andex[0][i] = manager->cache.andex[1][i] =
		manager->cache.andex[2][i] = 
		NULL;
   
   	/* allocate and initialize zero and one constants */
      	manager->constant.one = (mdd*) malloc(sizeof(mdd));
   	manager->constant.zero = (mdd*) malloc(sizeof(mdd));
   	manager->constant.one->code = manager->constant.zero->code = 0;
	manager->constant.one->low  = manager->constant.zero->low  =
	manager->constant.one->high = manager->constant.zero->high = 
	manager->constant.one->next = manager->constant.zero->next = NULL;
   	mdd_flag_set(manager->constant.zero,mdd_flag_zero);
	mdd_flag_set(manager->constant.one,mdd_flag_one);
	mdd_flag_set(manager->constant.zero,mdd_flag_lock);
	mdd_flag_set(manager->constant.one,mdd_flag_lock);

      	/* initialize user references list */
	mdd_array_alloc(mdd*,mdd_growth,manager->references.nodes);
	manager->references.index = 0;
	manager->references.size = mdd_growth;

	/* print manager informations 
	mdd_manager_print_info(manager,stderr);*/
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_create");
#endif
   	/* return this manager */
   	return manager;
   	}

void 
mdd_manager_destroy(manager)
mdd_manager* manager;
/*
Destroy a mdd manager.  Free all memory used by manager
hash table, manager caches and others arrays.
*/
	{
   	int i;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_destroy");
#endif
	/* delete node hash table, old dynamic allocation ...
	for(i=0;i<manager->hashtable.size;i++){
   		mdd *crt, *tmp;
   		crt = manager->hashtable.entry[i];
   		while (crt != NULL){
   			tmp = crt->next;
   			free(crt);
   			crt = tmp;
   			}
  		} */
	mdd_array_free(manager->hashtable.entry);
	/* free nodes */
	mdd_array_free(manager->table.nodes);
	/* delete caches */
	for(i=0;i<3;i++){
   		mdd_array_free(manager->cache.and[i]);
   		mdd_array_free(manager->cache.or[i]);
		mdd_array_free(manager->cache.andex[i]);
   		if (i!=2){
			mdd_array_free(manager->cache.not[i]);
  			mdd_array_free(manager->cache.exist[i]);
      			mdd_array_free(manager->cache.subst[i]);
      			}
      		}
	/* delete variable informations */
	mdd_array_free(manager->variable.range);
	mdd_array_free(manager->variable.buffer.low);
	mdd_array_free(manager->variable.buffer.high);

	/* delete user references list */
	mdd_array_free(manager->references.nodes);

#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_destroy");
#endif
   	}

void
mdd_manager_print_info(manager,file)
mdd_manager* manager;
FILE* file;
/*
Print some information about a mdd manager in a given file, 
memory used, hash table informations, cache sizes...
*/
	{
   	int i,sum,min=1000000,max=-1;
   	float var;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_print_info");
#endif
	fprintf(file,"\n- %d stored mdd nodes, %5.2f%% capacity\n",
		manager->hashtable.nodes,
		(float) manager->hashtable.nodes * 100.0 / (float) manager->hashtable.capacity);
	/* compute the hash table statistics */
	sum = 0;
	for(i=0; i<manager->hashtable.size;i++){
   		mdd* crt;int cnt;
   		for (cnt=0,crt=manager->hashtable.entry[i];
      			crt!=NULL;crt=crt->next,cnt++);
		if (cnt < min)
		        min = cnt;
		if (max < cnt)
		        max = cnt;
   		sum += cnt * cnt; 
   		}
	/* hash table variance */
   	var = ((float)sum / (float)manager->hashtable.size) - 
   		((float)manager->hashtable.nodes * manager->hashtable.nodes /
      		((float)manager->hashtable.size * manager->hashtable.size));
	fprintf(file,"- collisions %d min, %d max, %d avg, %.2f dev\n",
		min,max,manager->hashtable.nodes/manager->hashtable.size,
		sqrt((var * manager->hashtable.size)/(manager->hashtable.size - 1)));
        /* count user references */
        for(sum=0,i=0;i<manager->references.index;i++)
	        if (manager->references.nodes[i]!=NULL)
		        sum++;
        fprintf(file,"- %d active references\n\n",sum);
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_print_info");
#endif
   	}

/*
-------------------------------------------------------------
        MDD Nodes Basic Creation and Management 
-------------------------------------------------------------
*/


mdd*
mdd_one(manager)
mdd_manager* manager;
	{
   	return manager->constant.one;
   	}

mdd*
mdd_zero(manager)
mdd_manager* manager;
	{
   	return manager->constant.zero;
   	}

void
mdd_declare_variable(manager,index,range)
mdd_manager* manager;
int index,range;
/*
Declare a variable 'index' with a variable 'range'. The variable
index must not to be great  than the number of manager variables
and the range must be in the  interval  [2,256].   If a variable
with the  same index was  already declared, this declaration are 
ignored if the range are the same, else will be considered error.  
*/
	{
#ifdef mdd_trace
	mdd_message("trace : begin mdd_declare_variable");
#endif
	/* test the initial arguments */
	if (index < 0 || index > manager->variable.number - 1){
		mdd_message("mdd_declare_variable : invalid variable index");
   		mdd_abort();
  		}
	if (range < 2 || range > 256){
		mdd_message("mdd_declare_variable : invalid range size");
   		mdd_abort();
  		}

	/* test if variable was already declared */
	if (manager->variable.range[index] != 0 
   	&& manager->variable.range[index] != range){
		mdd_message("mdd_declare_variable : variable already declared");
		mdd_abort();
   		}
	else
		manager->variable.range[index] = range;
#ifdef mdd_trace
	mdd_message("trace : end mdd_declare_variable");
#endif
 	}

mdd*
mdd_get_variable(manager,index,value)
mdd_manager* manager;
int index, value;
/*
Create  and  return a mdd, for encode the basic relation 

		  v_index == value

The arguments are the 'index' and the 'value' of variable.  
The manager check the index, which must not be great than 
the  manager  number of variables,  check the declaration 
for the  variable with the same index and check the value, 
which  must  be  in the declared range.
*/
	{
   	mdd* res;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_get_variable");
#endif
	/* test the initial arguments */
	if (index < 0 || index > manager->variable.number - 1){
		mdd_message("mdd_get_variable : invalid variable index");
  		mdd_abort();
   		}
	if (manager->variable.range[index] == 0){
		mdd_message("mdd_get_variable : variable not declared");
   		mdd_abort();
  		}
	if (value < 0 || value > manager->variable.range[index] - 1){
		mdd_message("mdd_get_variable : invalid variable value");
		mdd_abort();
  		}

	/* construct the mdd, in three different cases  */
	if (value == 0)
		res = mdd_manager_search_or_insert(manager,index,
   			value,manager->constant.one,manager->constant.zero);
	else if (value == manager->variable.range[index] - 1)
   		res = mdd_manager_search_or_insert(manager,index,
      			value-1,manager->constant.zero,manager->constant.one);
   	else{
   		mdd* tmp = mdd_manager_search_or_insert(manager,index,
      			value,manager->constant.one,manager->constant.zero);
   		mdd_manager_keep(manager,tmp);
   		res = mdd_manager_search_or_insert(manager,index,value - 1,
      			manager->constant.zero,tmp);
   		mdd_manager_release(manager,tmp);
   		}

#ifdef mdd_trace
	mdd_message("trace : end mdd_get_variable");
#endif
	/* return the node */
	return res;
   	}


void
mdd_manager_keep(manager,x)
mdd_manager* manager;
mdd* x;
/*
Add the mdd 'x' to the list of user references, which contain 
the nodes  referred  by user pointers which will be protected 
during the garbage collection process.
*/
	{
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_keep");
#endif
   	if (!mdd_is_constant(x)){
		/* test the references array */
		if (manager->references.index==manager->references.size){
			int i, k;
			/* growth and recompact the references array */
			mdd** rfs = (mdd**)malloc(sizeof(mdd*)*(manager->references.size+mdd_growth));
			for(k=0,i=0;i<manager->references.size;i++)
			        if (manager->references.nodes[i]!=NULL)
				        rfs[k++]=manager->references.nodes[i];
			/* delete old references array */
			free(manager->references.nodes);
			/* manager references fields updating */
			manager->references.nodes = rfs;
			manager->references.size += mdd_growth;
			manager->references.index = k;
			/* debugging
			printf("\tmdd : references updating : %d/%d\n",
			       manager->references.index, manager->references.size);*/
		        }
		/* keep the reference */
		manager->references.nodes[manager->references.index++] = x;
      		}
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_keep");
#endif
   	}

void
mdd_manager_release(manager,x)
mdd_manager* manager;
mdd* x;
/*
Remove the mdd 'x' from the list of user references.
*/
	{
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_release");
#endif
   	if (!mdd_is_constant(x)){
	        /* last position test */
	        if (manager->references.index && manager->references.nodes[manager->references.index-1] == x)
		        manager->references.index--;
		else{ /* any other position */
		        int i;
			for(i=manager->references.index-2;i>-1;i--)
			        if (manager->references.nodes[i] == x){
				        manager->references.nodes[i] = NULL;
					break;
			                }
			if (i<0)
			        mdd_message("release : not a valid reference");
		        }
     		}
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_release");
#endif
   	}

/*
-------------------------------------------------------------
        Operations on MDD's formulas
-------------------------------------------------------------
*/

mdd*
mdd_not(manager,x)
mdd_manager* manager;
mdd* x;
/*
Recursively, perform the negation of a given mdd.
*/
	{
   	mdd* res;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_not");
#endif
	/* test the elementary cases */
	if (mdd_is_one(x))
		res = manager->constant.zero;
	else if (mdd_is_zero(x))
		res = manager->constant.one;
	/* the node are internal */
	else{
   		/* compute the cache entry */
   		int k = ((int)x >> 2) % manager->cache.size;
   		if (manager->cache.not[0][k] == x)
   			res = manager->cache.not[1][k];
   		else{
      			mdd *low,*high;
      			/* compute the negation for successors */
      			low = mdd_not(manager,x->low);
      			mdd_manager_keep(manager,low);
      			high = mdd_not(manager,x->high);
      			mdd_manager_keep(manager,high);
      			res = mdd_manager_search_or_insert(manager, 
         			mdd_get_index(x), mdd_get_value(x), low,high);
      			mdd_manager_release(manager,high);
      			mdd_manager_release(manager,low);
      			/*add the result to the not cache */
      			manager->cache.not[0][k] = x;
      			manager->cache.not[1][k] = res; 
      			}
   		}
#ifdef mdd_trace
	mdd_message("trace : end mdd_not");
#endif
	/* return the result */
	return res;
   	}

mdd*
mdd_and(manager,x1,x2)
mdd_manager* manager;
mdd *x1,*x2;
/*
Recursively, perform the and of two mdd's.
*/
	{
   	mdd* res;
#ifdef mdd_trace
        mdd_message("trace : begin mdd_and");
#endif
	/* test the elementary cases */
	if (mdd_is_one(x1) || mdd_is_zero(x2))
		res = x2;
	else if (mdd_is_one(x2) || mdd_is_zero(x1))
		res = x1;
	else if (x1 == x2)
		res = x1;
	/* both nodes are internals , check first in the and cache */
	else{
   		/* compute the cache entry */
   		int k = (((int)x1 >> 2) + ((int)x2 >> 2)) % manager->cache.size;
   		if ((manager->cache.and[0][k] == x1 && manager->cache.and[1][k] == x2)
      		||  (manager->cache.and[0][k] == x2 && manager->cache.and[1][k] == x1))
   			res = manager->cache.and[2][k];
   		else{
      			mdd *low,*high;
      			/* the result was not found in cache */
      			if ((mdd_get_index(x1) > mdd_get_index(x2))
         		|| (mdd_get_index(x1) == mdd_get_index(x2)
            		&&  mdd_get_value(x1) > mdd_get_value(x2))){
         			/* permute the arguments */
         			mdd* tmp;
         			tmp = x1;
         			x1 = x2;
         			x2 = tmp;
         			}
      			if (mdd_get_index(x1) < mdd_get_index(x2)){
         			/* different variables */
         			low = mdd_and(manager,x1->low,x2);
         			mdd_manager_keep(manager,low);
         			high = mdd_and(manager,x1->high,x2);
         			mdd_manager_keep(manager,high);
         			}
      			else if (mdd_get_value(x1) < mdd_get_value(x2)){
         			/* the same variable but different values */
         			low = mdd_and(manager,x1->low,x2->low);
         			mdd_manager_keep(manager,low);
         			high = mdd_and(manager,x1->high,x2);
         			mdd_manager_keep(manager,high);
         			}
      			else{
         			/* the same variable and the same value */
         			low = mdd_and(manager,x1->low,x2->low);
         			mdd_manager_keep(manager,low);
         			high = mdd_and(manager,x1->high,x2->high);
         			mdd_manager_keep(manager,high);
         			}
      			if ((high == low) || (!mdd_is_constant(high)
         		&& mdd_get_index(high) == mdd_get_index(x1)
         		&& low == high->low))
      				res = high;
      			else
      				res = mdd_manager_search_or_insert(manager,
         				mdd_get_index(x1),mdd_get_value(x1),low,high);
      			mdd_manager_release(manager,high);
      			mdd_manager_release(manager,low);
   			manager->cache.and[0][k] = x1;
   			manager->cache.and[1][k] = x2;
   			manager->cache.and[2][k] = res;
   			}
   		}
#ifdef mdd_trace
        mdd_message("trace : end mdd_and");
#endif
	/* return the and mdd */
	return res;
   	}


mdd*
mdd_or(manager,x1,x2)
mdd_manager* manager;
mdd *x1,*x2;
/*
Recursively, perform the or of two mdd's.
*/
        {
   	mdd* res;
#ifdef mdd_trace
        mdd_message("trace : begin mdd_or");
#endif
        /* test the elementary cases */
        if (mdd_is_one(x1) || mdd_is_zero(x2))
                res = x1;
        else if (mdd_is_one(x2) || mdd_is_zero(x1))
                res = x2;
	else if (x1 == x2)
		res = x1;
        /* both nodes are internals, check first in the or cache*/
        else{
                /* compute the cache entry */
                int k = (((int)x1 >> 2) + ((int)x2 >> 2)) % manager->cache.size;
                if ((manager->cache.or[0][k] == x1 && manager->cache.or[1][k] == x2)
                ||  (manager->cache.or[0][k] == x2 && manager->cache.or[1][k] == x1))
                        res = manager->cache.or[2][k];
                else{
                        /* the result was not found in cache */
                        mdd *low,*high;
      			if ((mdd_get_index(x1) > mdd_get_index(x2))
         		|| ( mdd_get_index(x1) == mdd_get_index(x2)
            		&&   mdd_get_value(x1) > mdd_get_value(x2))){
         			/* permute the arguments */
         			mdd* tmp;
         			tmp = x1;
         			x1 = x2;
         			x2 = tmp;
         			}
                        if (mdd_get_index(x1) < mdd_get_index(x2)){
         			/* different variables */
                                low = mdd_or(manager,x1->low,x2);
         			mdd_manager_keep(manager,low);
                                high = mdd_or(manager,x1->high,x2);
         			mdd_manager_keep(manager,high);
                                }
                        else if (mdd_get_value(x1) < mdd_get_value(x2)){
         			/* the same variable but different values */
                                low = mdd_or(manager,x1->low,x2->low);
         			mdd_manager_keep(manager,low);
                                high = mdd_or(manager,x1->high,x2);
         			mdd_manager_keep(manager,high);
                                }
                        else{
         			/* the same variable and the same value */
                                low = mdd_or(manager,x1->low,x2->low);
         			mdd_manager_keep(manager,low);
                                high = mdd_or(manager,x1->high, x2->high);
         			mdd_manager_keep(manager,high);
                                }
                        if ((high == low) || (!mdd_is_constant(high)
            		&& mdd_get_index(high) == mdd_get_index(x1)
            		&& low == high->low))
                                res = high;
                        else
                                res = mdd_manager_search_or_insert(manager,
         				mdd_get_index(x1),mdd_get_value(x1),low,high);
      
                        mdd_manager_release(manager,high);
                        mdd_manager_release(manager,low);
                	manager->cache.or[0][k] = x1;
                	manager->cache.or[1][k] = x2;
                	manager->cache.or[2][k] = res;
      			}
                }
#ifdef mdd_trace
        mdd_message("trace : end mdd_or");
#endif
	/* return the or mdd */
	return res;
        }

mdd*
mdd_exist(manager,x,vars)
mdd_manager* manager;
mdd* x;
int* vars;
/*
Compute the existential quantification, for some variables.
The quantified variables must have a non zero value, in the
vars array.  
*/
	{
   	mdd* res;
   	int i;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_exist");
#endif
	/* purge the manager exist cache */
	for(i=0;i<manager->cache.size;i++)
		manager->cache.exist[0][i] = manager->cache.exist[1][i] = 
		NULL;
	/* compute the quantification */
	res = mdd_basic_exist(manager,x,vars);
#ifdef mdd_trace
	mdd_message("trace : end mdd_exist");
#endif
	/* return the quantification */
	return res;
	}

mdd* 
mdd_basic_exist(manager,x,vars)
mdd_manager* manager;
mdd* x;
int* vars;
	{
   	mdd* res;
   	/* test the elementary case */
	if (mdd_is_constant(x))
		res = x;
   	/* the node is variable */
	else{
      		/* compute the exist cache entry */
      		int k = ((int)x >> 2) % manager->cache.size;
   		if (manager->cache.exist[0][k] == x)
   			res = manager->cache.exist[1][k];
   		else{
         		/* compute the quantification */
   			mdd *low,*high;
      			if (vars[mdd_get_index(x)] != 0){
         			/* the node variable is quantified */
   				low = mdd_basic_exist(manager,x->low,vars);
   				mdd_manager_keep(manager,low);
   				high = mdd_basic_exist(manager,x->high,vars);
   				mdd_manager_keep(manager,high);
         			/* compute the result */
         			res = mdd_or(manager,low,high);
        			}
      			else{
         			/* the node variable is not quantified */
         			low = mdd_basic_exist(manager,x->low,vars);
         			mdd_manager_keep(manager,low);
         			high = mdd_basic_exist(manager,x->high,vars);
         			mdd_manager_keep(manager,high);
         			/* compute the result */
				if (low == x->low && high == x->high)
				        res = x;
				else if ((low == high) || (!mdd_is_constant(high)
               			&& mdd_get_index(high) == mdd_get_index(x)
               			&& high->low == low))
         				res = high;
         			else
         				res = mdd_manager_search_or_insert(manager,
            					mdd_get_index(x),mdd_get_value(x),low,high);
         			} 		
   			mdd_manager_release(manager,high);
   			mdd_manager_release(manager,low);
      			/* add the result to the exist cache */
      			manager->cache.exist[0][k] = x;
      			manager->cache.exist[1][k] = res;
      			}
   		}
	/* return the final result */
	return res;
   	}


mdd*
mdd_and_exist(manager, x1, x2, vars)
mdd_manager* manager;
mdd *x1, *x2;
int* vars;
/*
Compute the and of two mdd's and apply also an existential 
quantification with some variables.
*/
	{
	mdd* res;
	int i;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_and_exist");
#endif
	/* purge the manager andexist and exist caches */
	for(i=0;i<manager->cache.size;i++)
		manager->cache.andex[0][i] = manager->cache.andex[1][i] =
		manager->cache.andex[2][i] = manager->cache.exist[0][i] =
		manager->cache.exist[1][i] = NULL;
	/* compute the result */
	res = mdd_basic_and_exist(manager,x1,x2,vars);
#ifdef mdd_trace
	mdd_message("trace : end mdd_and_exist");
#endif	
	/* return the final result */
	return res;
	}

mdd*
mdd_basic_and_exist(manager,x1,x2,vars)
mdd_manager* manager;
mdd *x1, *x2;
int* vars;
	{
	mdd* res;
	/* test the elementary cases */
	if (mdd_is_one(x1))
		res = mdd_basic_exist(manager, x2, vars);
	else if (mdd_is_one(x2))
		res = mdd_basic_exist(manager, x1, vars);
	else if (mdd_is_zero(x1) || mdd_is_zero(x2))
		res = manager->constant.zero;
	else if (x1 == x2)
		res = mdd_basic_exist(manager,x1,vars);
	/* both nodes are internals, check first in the and exist cache */
	else{
   		/* compute the cache entry */
   		int k = (((int)x1 >> 2) + ((int)x2 >> 2)) % manager->cache.size;
   		if ((manager->cache.andex[0][k] == x1 && manager->cache.andex[1][k] == x2)
      		||  (manager->cache.andex[0][k] == x2 && manager->cache.andex[1][k] == x1))
   			res = manager->cache.andex[2][k];
   		else{
      			mdd *low,*high;
      			/* the result was not found in cache */
      			if ((mdd_get_index(x1) > mdd_get_index(x2))
         		|| (mdd_get_index(x1) == mdd_get_index(x2)
            		&&  mdd_get_value(x1) > mdd_get_value(x2))){
         			/* permute the arguments */
         			mdd* tmp;
         			tmp = x1;
         			x1 = x2;
         			x2 = tmp;
         			}
			/* recursively perform the and exist */
      			if (mdd_get_index(x1) < mdd_get_index(x2)){
         			/* different variables */
         			low = mdd_basic_and_exist(manager,x1->low,x2,vars);
         			mdd_manager_keep(manager,low);
         			high = mdd_basic_and_exist(manager,x1->high,x2,vars);
         			mdd_manager_keep(manager,high);
         			}
      			else if (mdd_get_value(x1) < mdd_get_value(x2)){
         			/* the same variable but different values */
         			low = mdd_basic_and_exist(manager,x1->low,x2->low,vars);
         			mdd_manager_keep(manager,low);
         			high = mdd_basic_and_exist(manager,x1->high,x2,vars);
         			mdd_manager_keep(manager,high);
         			}
      			else{
         			/* the same variable and the same value */
         			low = mdd_basic_and_exist(manager,x1->low,x2->low,vars);
         			mdd_manager_keep(manager,low);
         			high = mdd_basic_and_exist(manager,x1->high,x2->high,vars);
         			mdd_manager_keep(manager,high);
         			}
			/* test the intermediary results */
      			if ((high == low) || (!mdd_is_constant(high)
         		&& mdd_get_index(high) == mdd_get_index(x1)
         		&& low == high->low))
				/* reduce case */
      				res = high;
      			else if (vars[mdd_get_index(x1)])
				/* crt variable x1 is quantified */
				res = mdd_or(manager, low, high);
			else
				/* crt variable x1 is not quantified  */
      				res = mdd_manager_search_or_insert(manager,
         				mdd_get_index(x1),mdd_get_value(x1),low,high);
      			mdd_manager_release(manager,high);
      			mdd_manager_release(manager,low);
   			manager->cache.andex[0][k] = x1;
   			manager->cache.andex[1][k] = x2;
   			manager->cache.andex[2][k] = res;
   			}
   		}
	return res;
	}

mdd*
mdd_substitute(manager,x,corresp)
mdd_manager* manager;
mdd* x;
int* corresp;
/*
Compute the substitution of variables using a substitution list.
First, it check the substitution list, for the range equivalence.
*/
        {
   	mdd* res;
   	int i;
#ifdef mdd_trace
        mdd_message("trace : begin mdd_substitute");
#endif
	/* check variables range */
	for(i=0;i<manager->variable.number;i++)
		if (manager->variable.range[i] != manager->variable.range[corresp[i]]){
   			mdd_message("mdd_substitute : invalid substitution list");
   			mdd_abort();
   			}
	/* purge the manager substitution cache */
	for(i=0;i<manager->cache.size;i++)
		manager->cache.subst[0][i] = manager->cache.subst[1][i] =
		NULL;
	/* compute the substitution */
	res = mdd_basic_substitute(manager,x,corresp);
#ifdef mdd_trace
	mdd_message("trace : end mdd_substitute");
#endif
	/* return the substitution */
	return res;
	}

mdd*
mdd_basic_substitute(manager,x,corresp)
mdd_manager* manager;
mdd* x;
int* corresp;
	{
   	mdd* res;
	/* test the elementary case */
	if (mdd_is_constant(x))
		res = x;
   	/* the node is variable */
	else{
      		/* compute the subst cache entry */
      		int k = ((int) x >> 2) % manager->cache.size;
      		if (manager->cache.subst[0][k] == x)
      			res = manager->cache.subst[1][k];
      		else {
   			mdd *low,*high;
   			/* compute the substitution for low and high branches */
   			low = mdd_basic_substitute(manager,x->low,corresp);
   			mdd_manager_keep(manager,low);
   			high = mdd_basic_substitute(manager,x->high,corresp);
   			mdd_manager_keep(manager,high);
   			if ((high == low) || (!mdd_is_constant(high)
            		&& mdd_get_index(high) == corresp[mdd_get_index(x)]
            		&& high->low == low))
   				res = high;
   			else
   				res = mdd_manager_search_or_insert(manager,
      					corresp[mdd_get_index(x)],mdd_get_value(x),low,high);
   			mdd_manager_release(manager,high);
   			mdd_manager_release(manager,low);
         		/* add the result to the subst cache */
         		manager->cache.subst[0][k] = x;
         		manager->cache.subst[1][k] = res;
        		}
   		}
	/* return the substitution */
	return res;
        }

void
mdd_print(manager,x,file)
mdd_manager* manager;
mdd* x;
FILE* file;
/*
Print the mdd internal structure in a file.
*/
        {
#ifdef mdd_trace
        mdd_message("trace : begin mdd_print");
#endif
	/* first mark all mdd internal nodes */
	mdd_flag_recset(x,mdd_flag_mark);
	/* now, print `one time' all marked mdd nodes */
	mdd_basic_print(manager,x,file);
#ifdef mdd_trace
        mdd_message("trace : end mdd_print");
#endif
        }

void mdd_basic_print(manager,x,file)
mdd_manager* manager;
mdd* x;
FILE* file;
	{
   	/* test the elementary cases */
   	if (mdd_is_one(x))
   		fprintf(file," One\n");
   	else if (mdd_is_zero(x))
   		fprintf(file," Zero\n");
 	/* only, if the node was not yet printed, print it */
   	else if (mdd_flag_test(x,mdd_flag_mark)){
      		/* first, print the successors, if they are not constants */
      		if (!mdd_is_constant(x->low))
      			mdd_basic_print(manager,x->low,file);
      		if (!mdd_is_constant(x->high))
      			mdd_basic_print(manager,x->high,file);
      		/* print this node */
      		fprintf(file,"%#x : <If> (x%-6d<=%3d) ",
         		x,mdd_get_index(x),mdd_get_value(x));
      		if (mdd_is_constant(x->low))
      			fprintf(file,"<Then> %8s ",mdd_is_one(x->low) ? "One" : "Zero");
      		else
      			fprintf(file,"<Then> %#x ",x->low);
      		if (mdd_is_constant(x->high))
      			fprintf(file,"<Else> %8s\n",mdd_is_one(x->high) ? "One" : "Zero");
      		else
      			fprintf(file,"<Else> %#x\n",x->high);
      		/* mark current node as printed */
      		mdd_flag_reset(x,mdd_flag_mark);	
      		}
   	}

int
mdd_size(manager,x)
mdd_manager* manager;
mdd* x;
/*
Compute the size of mdd, the number of internal nodes.
*/
        {
   	int res = 0;
#ifdef mdd_trace
        mdd_message("trace : begin mdd_size");
#endif
	/* first mark all mdd internal nodes */
	mdd_flag_recset(x,mdd_flag_mark);
	/* count the marked nodes */
	mdd_basic_size(manager,x,&res);
#ifdef mdd_trace
        mdd_message("trace : end mdd_size");
#endif
	/* return the number of nodes */
	return res;
        }

void
mdd_basic_size(manager,x,counter)
mdd_manager* manager;
mdd* x;
int* counter;
	{
  	/* if the node is not counted and it is not constant */
   	if (!mdd_is_constant(x) && mdd_flag_test(x,mdd_flag_mark)){
      		/* count the successors */
      		mdd_basic_size(manager,x->low,counter);
      		mdd_basic_size(manager,x->high,counter);
      		/* increment counter and reset node mark */
      		mdd_flag_reset(x,mdd_flag_mark);
      		(*counter)++;
      		}
   	}

void
mdd_dump(manager,x,file)
mdd_manager* manager;
mdd* x;
FILE* file;
/*
Enumerate in a file the n-uples encoded by the mdd.
*/
        {
#ifdef mdd_trace
        mdd_message("trace : begin mdd_enumerate");
#endif
	mdd_basic_dump(manager,x,file,-1);
#ifdef mdd_trace
        mdd_message("trace : end mdd_enumerate");
#endif
        }

void
mdd_basic_dump(manager,x,file,last_index)
mdd_manager* manager;
mdd* x;
FILE* file;
int last_index;
	{
	int i;
	/* elementary case, dump manager values buffer */
	if (mdd_is_one(x)) {
	        for(i=last_index+1;i<manager->variable.number;i++) {
	                manager->variable.buffer.low[i] = 0;
		        manager->variable.buffer.high[i] = -1;
	                }
	        for(i=0;i<manager->variable.number;i++)
		if (manager->variable.range[i] 
		&&  manager->variable.buffer.low[i] <= manager->variable.buffer.high[i]) {
		        if (manager->variable.buffer.low[i] == manager->variable.buffer.high[i])
			        fprintf(file,"x%d[%d] ",i,manager->variable.buffer.low[i]);
			else
			        fprintf(file,"x%d[%d,%d] ",i,manager->variable.buffer.low[i],
					manager->variable.buffer.high[i]);
		        }
		fprintf(file,"\n");
	        }
	/* dump all low branches for current variable index */
	else if (! mdd_is_zero(x)) {
	        mdd* crt;
	        for(i=last_index+1;i<mdd_get_index(x);i++) {
	                manager->variable.buffer.low[i] = 0;
		        manager->variable.buffer.high[i] = -1;
	                }
		i = mdd_get_index(x);
		crt = x;
		manager->variable.buffer.high[i] = -1;
		while (i==mdd_get_index(crt) && !mdd_is_constant(crt)) {
		        /* update buffer values */
		        manager->variable.buffer.low[i] = manager->variable.buffer.high[i]+1;
			manager->variable.buffer.high[i] = mdd_get_value(crt);
			/* dump crt low branch */
			mdd_basic_dump(manager, crt->low, file, i);
			/* go to next node */
			crt = crt->high;
		        }
		/* dump the final high branch */
		manager->variable.buffer.low[i] = manager->variable.buffer.high[i]+1;
		manager->variable.buffer.high[i] = manager->variable.range[i] - 1;
		mdd_basic_dump(manager, crt, file, i);
	        }
      	}


double
mdd_onset_size(manager,x,onset)
mdd_manager* manager;
mdd* x;
int* onset;
/*
Compute the number of n-uples encoded by mdd, which
is supposed to be defined only on onset domain.
*/
        {
   	double res;
#ifdef mdd_trace
        mdd_message("trace : begin mdd_onset_size");
#endif
	res = mdd_basic_onset_size(manager,x,onset,-1);
#ifdef mdd_trace
        mdd_message("trace : end mdd_onset_size");
#endif
	return res;
        }

double
mdd_basic_onset_size(manager,x,onset,last_index)
mdd_manager* manager;
mdd* x;
int* onset;
int last_index;
	{
   	double res = 0.0;
   	int i;
   	/* if the node is 'one' count the n-uples */
   	if (mdd_is_one(x)){
      		/* the number is the amplification factor */
      		for(res=1.0,i=last_index+1;i<manager->variable.number;i++)
      		if (manager->variable.range[i] && onset[i])
      			res *= (double)(manager->variable.range[i]);
      		}
   	else if (! mdd_is_zero(x)){
      		int last_value;
      		double fct, cnt;
      		mdd* crt;
      		/* test if the variable node is onset */
      		if (onset[mdd_get_index(x)] == 0){
         		mdd_message("mdd_onset_size : unexpected variable index");
         		mdd_abort();
         		}
      		/* compute the amplification factor */
      		for(fct=1.0,i=last_index+1;i<mdd_get_index(x);i++)
      		if (manager->variable.range[i] && onset[i])
      			fct *= (double)(manager->variable.range[i]);
      		/* count n-uplets for low branches for all node with the 
      		   same variable index, and after, the final high branch */
      		last_value = -1;
      		crt = x;
      		while (mdd_get_index(crt) == mdd_get_index(x) && !mdd_is_constant(crt)){
         		/* count the nodes for the low branch of crt */
         		cnt = mdd_basic_onset_size(manager,crt->low,onset,mdd_get_index(x));
         		/* add this number to the general result */
         		res += (cnt * (double)(mdd_get_value(crt) - last_value));
         		/* go to the next node and change the last value */
         		last_value = mdd_get_value(crt);
         		crt = crt->high;
         		}
      		/* count n-uples for the final crt node */
      		cnt = mdd_basic_onset_size(manager,crt,onset,mdd_get_index(x));
      		/* also, add this result to the general result */
      		res += (cnt * (double)(manager->variable.range[mdd_get_index(x)] - last_value -1));
      		/* final, amplif the result with fct */
      		res *= fct;
      		}
   	/*printf("trace %#x ----> %.0lf",x,res);*/
   	return res;
   	}


/*
-------------------------------------------------------------
        MDD Local Functions
-------------------------------------------------------------
*/

mdd*
mdd_manager_search_or_insert(manager,index,value,low,high)
mdd_manager* manager;
int index,value;
mdd* low, *high;
/*
Search or Insert the mdd nodes in the manager hash table. If the 
mdd already exist in table, it is returned, if not, and if exist 
memory ,  the node is created , initialized and  inserted in the 
manager hash table. 
*/
	{
   	mdd* crt;
   	int hash;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_search_or_insert");
#endif
   	/* compute the hash code entry */
   	hash = (((int)low >> 4) + ((int)high >> 4) + ((index + 1) << 5) + 
      		((value + 1) << 7)) % manager->hashtable.size;
   	/* search the node in hash table */
   	for (crt=manager->hashtable.entry[hash];crt!=NULL;crt=crt->next)
   		if (crt->low == low && crt->high == high 
      		&&  mdd_get_index(crt) == index 
      		&&  mdd_get_value(crt) == value)
   			break;
   	/* if it is not found, create and insert a new node */
   	if (crt == NULL){
      		/* check for existing table space */
      		if (manager->hashtable.capacity == manager->hashtable.nodes
         	&& !mdd_manager_garbage_collection(manager)){
      			mdd_message("mdd_manager_search_or_insert : insufficient memory (1)");
      			mdd_abort();
      			}
      		else{
		        /* the old dynamic allocation ... 
		        if ((crt = (mdd*) malloc(sizeof(mdd)))==NULL) {
			        mdd_message("mdd_manager_search_or_insert : insufficient memory (2)");
			        mdd_abort();
		                }  */
		        /* check an existing not busy mdd node */
			while (mdd_flag_test(manager->table.crt, mdd_flag_busy))
			        manager->table.crt++;
			crt = manager->table.crt;
         		/* initialize node fields */
         		crt->code = 0;
			mdd_flag_set(crt,mdd_flag_busy);
         		mdd_set_index(crt,index);
         		mdd_set_value(crt,value);
        		crt->low = low;
         		crt->high = high;
         		crt->next = manager->hashtable.entry[hash];
         		/* insert the node in manager table */
         		manager->hashtable.entry[hash] = crt;
         		manager->hashtable.nodes++;
         		}
     		}
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_search_or_insert");
#endif
   	/* return the right node */
   	return crt;
   	}

int
mdd_manager_garbage_collection(manager)
mdd_manager* manager;
/*
Perform a manager garbage collection, all hash table nodes which 
are not protected from the user references list are deleted. Also, 
all caches invalid entries are reinitialized with NULL values...
*/
	{
   	int i, k;
#ifdef mdd_trace
	mdd_message("trace : begin mdd_manager_garbage_collection");
#endif
	fprintf(stderr,"(mdd garbage collection... ");
   	fflush(stderr);
   	/* lock user mdd's and recompact the list */
	for(i=0, k=0;i<manager->references.index;i++)
	         if (manager->references.nodes[i] != NULL){
		         manager->references.nodes[k++] = manager->references.nodes[i];
			 mdd_flag_recset(manager->references.nodes[i],mdd_flag_lock);
		         }
	manager->references.index = k;
	/* reinitialize caches, test and keep all valid entries */
	for(i=0;i<manager->cache.size;i++){
	        if (manager->cache.and[0][i]!=NULL 
		&& !(mdd_flag_test(manager->cache.and[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.and[1][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.and[2][i],mdd_flag_lock)))
		        manager->cache.and[0][i] = 
			manager->cache.and[1][i] =
   		        manager->cache.and[2][i] = NULL;
	        if (manager->cache.or[0][i]!=NULL
		&& !(mdd_flag_test(manager->cache.or[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.or[1][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.or[2][i],mdd_flag_lock)))
		        manager->cache.or[0][i] = 
			manager->cache.or[1][i] =
   		        manager->cache.or[2][i] = NULL;
	        if (manager->cache.andex[0][i]!=NULL
		&& !(mdd_flag_test(manager->cache.andex[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.andex[1][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.andex[2][i],mdd_flag_lock)))
		        manager->cache.andex[0][i] = 
			manager->cache.andex[1][i] =
   		        manager->cache.andex[2][i] = NULL;
	        if (manager->cache.not[0][i]!=NULL
		&& !(mdd_flag_test(manager->cache.not[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.not[1][i],mdd_flag_lock)))
		        manager->cache.not[0][i] = 
			manager->cache.not[1][i] = NULL;
	        if (manager->cache.subst[0][i]!=NULL
		&& !(mdd_flag_test(manager->cache.subst[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.subst[1][i],mdd_flag_lock)))
		        manager->cache.subst[0][i] = 
			manager->cache.subst[1][i] = NULL;
	        if (manager->cache.exist[0][i]!=NULL
		&& !(mdd_flag_test(manager->cache.exist[0][i],mdd_flag_lock)
		&& mdd_flag_test(manager->cache.exist[1][i],mdd_flag_lock)))
		        manager->cache.exist[0][i] = 
			manager->cache.exist[1][i] = NULL;
	        }
   	/* delete all unlocked mdd's */
   	for(i=0; i<manager->hashtable.size; i++){
   		mdd* crt;
      		crt = manager->hashtable.entry[i];
   		manager->hashtable.entry[i] = NULL;
   		while (crt != NULL){
      			mdd* next = crt->next;
      			if (mdd_flag_test(crt,mdd_flag_lock)){
         			mdd_flag_reset(crt,mdd_flag_lock);
         			crt->next = manager->hashtable.entry[i];
         			manager->hashtable.entry[i] = crt;
         			}
      			else{
			        /* the old dynamic allocation ...
				free(crt); */
			        mdd_flag_reset(crt, mdd_flag_busy);
         			manager->hashtable.nodes--;
         			}
      			crt = next;
      			}
   		}
	/* reinitialize current table node */
	manager->table.crt = manager->table.nodes;

	/* print some informations */
   	fprintf(stderr,"%5.2f %% done, refs %d/%d)\n",100.0 * (1.0 - 
      		(float)manager->hashtable.nodes/(float)manager->hashtable.capacity),
		manager->references.index, manager->references.size);
#ifdef mdd_trace
	mdd_message("trace : end mdd_manager_garbage_collection");
#endif
   	/* return the number of deleted nodes */
   	return (manager->hashtable.capacity - manager->hashtable.nodes);
   	}

void mdd_flag_recset(x,flag)
mdd* x;
int flag;
/*
Recursively, set the flag for the node 'x' and for all of its
successors.  This function is used to lock user mdd's for the
garbage collection and to mark the nodes for a given mdd, for
printing or counting.
*/
	{
   	/* test if the node has already the flag set or is constant */
   	if (!mdd_flag_test(x,flag) && !mdd_is_constant(x)){
      		/* set the flag for the node*/
      		mdd_flag_set(x,flag);
      		/* recursively, set the flag for successors */
		mdd_flag_recset(x->low,flag);
      		mdd_flag_recset(x->high,flag);      		
      		}
   	}

int 
mdd_next_prime(start)
int start;
/*
Give the next prime number after the start value. Used to 
compute the hash table and the cache size.
*/
   	{
   	if (start < 2 || start==2)
      		return 2;
   	else{
      		int p,k;
      		for(p=start;;p++)
         	if (p % 2)
         	for(k=3;;k+=2){
             		if (p / k < k)
                		return p;
             		if (! (p % k))
                		break;
                	}
            	}   
   	}
